@@ -14,7 +14,7 @@ from ast import literal_eval
 from datetime import datetime
 from decimal import Decimal
 from functools import partial, partialmethod
-from math import log10, floor, fmod
+from math import floor, fmod
 import numpy as np
 import os.path
 import pandas as pd
@@ -41,11 +41,13 @@ draw_line_color = '#000000'
 draw_done_color = '#555555'
 significant_decimals = 8
 significant_eps = 1e-8
-max_zoom_points = 20 # number of visible candles at maximum zoom
+max_zoom_points = 20 # number of visible candles when maximum zoomed in
 top_graph_scale = 2
 clamp_grid = True
+right_margin_candles = 5 # whitespace at the right-hand side
 lod_candles = 3000
 lod_labels = 700
+cache_candle_factor = 3 # factor extra candles rendered to buffer
 y_label_width = 65
 
 windows = [] # no gc
@@ -54,6 +56,7 @@ sounds = {} # no gc
 plotdf2df = {} # for pandas df.plot
 epoch_period = 1e30
 last_ax = None # always assume we want to plot in the last axis, unless explicitly specified
+overlay_axs = [] # for keeping track of candlesticks in overlays
 viewrestore = False
 
 
@@ -72,12 +75,46 @@ class EpochAxisItem(pg.AxisItem):
 
 
 
+class YAxisItem(pg.AxisItem):
+    def __init__(self, vb, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.vb = vb
+
+    def tickStrings(self, values, scale, spacing):
+        return ['%g'%self.vb.yscale.xform(value) for value in values]
+
+
+
+class YScale:
+    def __init__(self, scaletype, scalef):
+        self.scaletype = scaletype
+        self.scalef = scalef
+
+    def set_scale(self, scale):
+        self.scalef = scale if self.scaletype != 'log' else 1
+
+    def xform(self, y):
+        if self.scaletype == 'log':
+            y = 10**y
+        return y * self.scalef
+
+    def invxform(self, y, verify=False):
+        if self.scaletype == 'log':
+            if verify and y <= 0:
+                return -1e6
+            y = np.log10(y)
+        else:
+            y /= self.scalef
+        return y
+
+
+
 class PandasDataSource:
     '''Candle sticks: create with five columns: time, open, close, hi, lo - in that order.
        Volume bars: create with three columns: time, open, close, volume - in that order.
        For all other types, time needs to be first, usually followed by one or more Y-columns.'''
     def __init__(self, df):
-        if type(df.index) == pd.DatetimeIndex:
+        if type(df.index) == pd.DatetimeIndex or df.index[0]>1e10:
             df = df.reset_index()
         self.df = df.copy()
         # manage time column
@@ -98,7 +135,7 @@ class PandasDataSource:
     @property
     def period(self):
         timecol = self.df.columns[0]
-        return self.df[timecol].diff().median()
+        return self.df[timecol].diff().median() / 1000
 
     @property
     def index(self):
@@ -119,6 +156,10 @@ class PandasDataSource:
         col = self.df.columns[self.col_data_offset+1]
         return self.df[col]
 
+    @property
+    def xlen(self):
+        return len(self.df)+right_margin_candles
+
     def calc_significant_decimals(self):
         absdiff = self.y.diff().abs()
         absdiff[absdiff<1e-30] = 1e30
@@ -129,14 +170,16 @@ class PandasDataSource:
         return decimals, smallest_diff
 
     def update_init_x(self, init_steps):
-        self.init_x0 = max(len(self.df)-init_steps, 0) - 0.5
-        self.init_x1 = len(self.df) - 0.5
+        self.init_x0 = max(self.xlen-init_steps, 0) - 0.5
+        self.init_x1 = self.xlen - 0.5
 
     def closest_time(self, x):
-        return self.df.loc[int(x), 0]
+        timecol = self.df.columns[0]
+        return self.df.loc[int(x), timecol]
 
     def addcols(self, datasrc):
-        self.scale_cols += [c+len(self.df.columns)-datasrc.col_data_offset for c in datasrc.scale_cols]
+        new_scale_cols = [c+len(self.df.columns)-datasrc.col_data_offset for c in datasrc.scale_cols]
+        self.scale_cols += new_scale_cols
         orig_col_data_cnt = len(self.df.columns)
         if _has_timecol(datasrc.df):
             timecol = self.df.columns[0]
@@ -161,6 +204,7 @@ class PandasDataSource:
         datasrc.init_x0 = self.init_x0
         datasrc.init_x1 = self.init_x1
         datasrc.col_data_offset = orig_col_data_cnt
+        datasrc.scale_cols = new_scale_cols
         self.cache_hilo_query = ''
         self.cache_hilo_answer = None
 
@@ -175,11 +219,14 @@ class PandasDataSource:
                 data[col] = df[col]
         data = data.reset_index()
         self.df = data[orig_cols]
-        self.init_x1 = len(self.df) - 0.5
+        self.init_x1 = self.xlen - 0.5
 
     def hilo(self, x0, x1):
         '''Return five values in time range: t0, t1, highest, lowest, number of rows.'''
-        x0,x1 = int(x0),int(x1)
+        if x0 == x1:
+            x0 = x1 = int(x1)
+        else:
+            x0,x1 = int(x0+0.5),int(x1)
         query = '%i,%i' % (x0,x1)
         if query != self.cache_hilo_query:
             self.cache_hilo_query = query
@@ -209,11 +256,11 @@ class PandasDataSource:
         colcnt -= 1 # time is always implied
         colidxs = [0] + list(range(self.col_data_offset, self.col_data_offset+colcnt))
         dfr = df.iloc[:,colidxs]
-        if yscale == 'log':
+        if yscale.scaletype == 'log' or yscale.scalef != 1:
             dfr = dfr.copy()
             for i in range(1, colcnt+1):
                 if dfr.iloc[:,i].dtype != object:
-                    dfr.iloc[:,i] = np.log10(dfr.iloc[:,i])
+                    dfr.iloc[:,i] = yscale.invxform(dfr.iloc[:,i])
         return dfr
 
     def __eq__(self, other):
@@ -288,10 +335,9 @@ class FinCrossHair:
         self.hline.setPos(y)
         self.xtext.setPos(x, y)
         self.ytext.setPos(x, y)
-        xtext = _epoch2local(self.ax.vb.datasrc, x, clamp_grid)
+        xtext = _epoch2local(self.ax.vb.datasrc, x)
         linear_y = y
-        if self.ax.vb.yscale == 'log':
-            y = 10**y
+        y = self.ax.vb.yscale.xform(y)
         rng = self.ax.vb.y_max - self.ax.vb.y_min
         rngmax = abs(self.ax.vb.y_min) + rng # any approximation is fine
         sd,se = (self.ax.significant_decimals,self.ax.significant_eps) if clamp_grid else (significant_decimals,significant_eps)
@@ -300,31 +346,31 @@ class FinCrossHair:
         far_bottom = self.ax.viewRect().y() + self.ax.viewRect().height()*0.1
         close2right = x > far_right
         close2bottom = linear_y < far_bottom
+        try:
+            for info in self.infos:
+                xtext,ytext = info(x,y,xtext,ytext)
+        except Exception as e:
+            print(e)
         space = '      '
         if close2right:
-            rxtext = xtext + space
-            rytext = ytext + space
+            xtext = xtext + space
+            ytext = ytext + space
             xanchor = [1,1]
             yanchor = [1,0]
         else:
-            rxtext = space + xtext
-            rytext = space + ytext
+            xtext = space + xtext
+            ytext = space + ytext
             xanchor = [0,1]
             yanchor = [0,0]
         if close2bottom:
-            rytext = ytext + space
+            ytext = ytext + space
             yanchor = [1,1]
             if close2right:
                 xanchor = [1,2]
         self.xtext.setAnchor(xanchor)
         self.ytext.setAnchor(yanchor)
-        try:
-            for info in self.infos:
-                rxtext,rytext = info(self.ax,x,y,rxtext,rytext)
-        except Exception as e:
-            print(e)
-        self.xtext.setText(rxtext)
-        self.ytext.setText(rytext)
+        self.xtext.setText(xtext)
+        self.ytext.setText(ytext)
 
 
 
@@ -420,16 +466,16 @@ class FinEllipse(pg.EllipseROI):
 
 
 class FinViewBox(pg.ViewBox):
-    def __init__(self, win, init_steps=300, yscale='linear', v_zoom_scale=1, *args, **kwargs):
+    def __init__(self, win, init_steps=300, yscale=YScale('linear', 1), v_zoom_scale=1, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.win = win
         self.yscale = yscale
         self.v_zoom_scale = v_zoom_scale
+        self.v_zoom_baseline = 0.5
         self.v_autozoom = True
         self.y_max = 1000
         self.y_min = 0
         self.y_positive = True
-        self.y_zoom_offset = 0.5
         self.force_range_update = 0
         self.rois = []
         self.draw_line = None
@@ -443,7 +489,6 @@ class FinViewBox(pg.ViewBox):
         if not self.datasrc:
             return
         datasrc.update_init_x(self.init_steps)
-        self.update_y_zoom(datasrc.init_x0, datasrc.init_x1)
 
     def wheelEvent(self, ev, axis=None):
         if ev.modifiers() == QtCore.Qt.ControlModifier:
@@ -527,6 +572,9 @@ class FinViewBox(pg.ViewBox):
         ev.accept()
 
     def mouseClickEvent(self, ev):
+        if _mouse_clicked(self, ev):
+            ev.accept()
+            return
         if ev.button() != QtCore.Qt.LeftButton or ev.modifiers() != QtCore.Qt.ControlModifier or not self.draw_line:
             return super().mouseClickEvent(ev)
         # add another segment to the currently drawn line
@@ -539,8 +587,8 @@ class FinViewBox(pg.ViewBox):
     def keyPressEvent(self, ev):
         if _key_pressed(self, ev):
             ev.accept()
-        else:
-            super().keyPressEvent(ev)
+            return
+        super().keyPressEvent(ev)
 
     def linkedViewChanged(self, view, axis):
         if not self.datasrc:
@@ -567,7 +615,7 @@ class FinViewBox(pg.ViewBox):
         tr = self.targetRect()
         x1 = tr.right() + steps
         startx = -0.5
-        endx = len(self.datasrc.x) - 0.5
+        endx = self.datasrc.xlen - 0.5
         if x1 > endx:
             x1 = endx
         x0 = x1 - tr.width()
@@ -594,7 +642,7 @@ class FinViewBox(pg.ViewBox):
             x1 = tr.right()
         # make edges rigid
         xl = max(round(x0-0.5)+0.5, -0.5)
-        xr = min(round(x1-0.5)+0.5, len(self.datasrc.df)-0.5)
+        xr = min(round(x1-0.5)+0.5, self.datasrc.xlen-0.5)
         dxl = xl-x0
         dxr = xr-x1
         if dxl > 0:
@@ -602,7 +650,7 @@ class FinViewBox(pg.ViewBox):
         if dxr < 0:
             x0 += dxr
         x0 = max(round(x0-0.5)+0.5, -0.5)
-        x1 = min(round(x1-0.5)+0.5, len(self.datasrc.df)-0.5)
+        x1 = min(round(x1-0.5)+0.5, self.datasrc.xlen-0.5)
         # fetch hi-lo and set range
         t0,t1,hi,lo,cnt = self.datasrc.hilo(x0, x1)
         if cnt < max_zoom_points:
@@ -611,12 +659,20 @@ class FinViewBox(pg.ViewBox):
             tr = self.viewRect()
             hi = tr.bottom()
             lo = tr.top()
-        rng = (hi-lo) / self.v_zoom_scale
-        rng = max(rng, 2e-7) # some very weird bug where too small scale stops rendering
-        base = (hi+lo)*self.y_zoom_offset
-        hi = base + rng*(1-self.y_zoom_offset)
-        lo = base - rng*self.y_zoom_offset
-        self.set_range(x0, lo, x1, hi)
+        if self.yscale.scaletype == 'log':
+            lo = max(1e-100, lo)
+            rng = (hi / lo) ** (1/self.v_zoom_scale)
+            rng = min(rng, 1e50) # avoid float overflow
+            base = (hi*lo) ** self.v_zoom_baseline
+            y0 = base / rng**self.v_zoom_baseline
+            y1 = base * rng**(1-self.v_zoom_baseline)
+        else:
+            rng = (hi-lo) / self.v_zoom_scale
+            rng = max(rng, 2e-7) # some very weird bug where high/low exponents stops rendering
+            base = (hi+lo) * self.v_zoom_baseline
+            y0 = base - rng*self.v_zoom_baseline
+            y1 = base + rng*(1-self.v_zoom_baseline)
+        self.set_range(x0, y0, x1, y1)
 
     def set_range(self, x0, y0, x1, y1):
         if x0 is None or x1 is None:
@@ -625,10 +681,9 @@ class FinViewBox(pg.ViewBox):
             x1 = tr.right()
         if np.isnan(y0) or np.isnan(y1):
             return
-        if self.yscale == 'log':
-            y0 = np.log10(y0) if y0 > 0 else -1
-            y1 = np.log10(y1) if y1 > 0 else -1
-        self.setRange(QtCore.QRectF(pg.Point(x0, y0), pg.Point(x1, y1)), padding=0)
+        _y0 = self.yscale.invxform(y0, verify=True)
+        _y1 = self.yscale.invxform(y1, verify=True)
+        self.setRange(QtCore.QRectF(pg.Point(x0, _y0), pg.Point(x1, _y1)), padding=0)
 
     def remove_last_roi(self):
         if self.rois:
@@ -683,20 +738,19 @@ class FinPlotItem(pg.GraphicsObject):
         self.paint(self.painter)
 
     def paint(self, p, *args):
-        viewRect = self.viewRect()
-        self.update_dirty_picture(viewRect)
+        self.update_dirty_picture(self.viewRect())
         p.drawPicture(0, 0, self.picture)
 
     def update_dirty_picture(self, visibleRect):
         if self.dirty or \
             visibleRect.left() < self.cachedRect.left() or \
             visibleRect.right() > self.cachedRect.right() or \
-            visibleRect.width() < self.cachedRect.width() / 3: # optimize when zooming in
+            visibleRect.width() < self.cachedRect.width() / cache_candle_factor: # optimize when zooming in
             self._generate_picture(visibleRect)
 
     def _generate_picture(self, boundingRect):
         w = boundingRect.width()
-        self.cachedRect = QtCore.QRectF(boundingRect.left()-w, 0, 3*w, 0)
+        self.cachedRect = QtCore.QRectF(boundingRect.left()-w, 0, cache_candle_factor*w, 0)
         self.generate_picture(self.cachedRect)
         self.dirty = False
 
@@ -723,6 +777,7 @@ class CandlestickItem(FinPlotItem):
         self.draw_shadow = draw_shadow
         self.candle_width = candle_width
         self.colorfunc = colorfunc
+        self.x_offset = 0
         super().__init__(ax, datasrc)
 
     def generate_picture(self, boundingRect):
@@ -736,6 +791,8 @@ class CandlestickItem(FinPlotItem):
         for shadow,frame,body,df_rows in self.colorfunc(self, self.datasrc, df):
             idxs = df_rows.index
             rows = df_rows.values
+            if self.x_offset:
+                idxs += self.x_offset
             if self.draw_shadow:
                 p.setPen(pg.mkPen(shadow))
                 for x,(t,open,close,high,low) in zip(idxs, rows):
@@ -811,14 +868,14 @@ def create_plot(title='Finance Plot', rows=1, init_zoom_periods=1e10, maximize=T
         win.showMaximized()
     # normally first graph is of higher significance, so enlarge
     win.ci.layout.setRowStretchFactor(0, top_graph_scale)
-    win.ci.setContentsMargins(0, 0, 0 ,0)
-    win.ci.setSpacing(0)
+    win.ci.setContentsMargins(0, 0, 0, 0)
+    win.ci.setSpacing(-1)
     axs = []
     prev_ax = None
     for n in range(rows):
-        try: ysc = yscale[n]
-        except: ysc = yscale
-        v_zoom_scale = 1 if ysc == 'log' else 0.97
+        ysc = yscale[n] if type(yscale) in (list,tuple) else yscale
+        ysc = YScale(ysc, 1)
+        v_zoom_scale = 0.97
         viewbox = FinViewBox(win, init_steps=init_zoom_periods, yscale=ysc, v_zoom_scale=v_zoom_scale)
         ax = prev_ax = _add_timestamp_plot(win, prev_ax, viewbox=viewbox, index=n, yscale=ysc)
         _set_plot_x_axis_leader(ax)
@@ -869,37 +926,32 @@ def candlestick_ochl(datasrc, draw_body=True, draw_shadow=True, candle_width=0.6
     datasrc.scale_cols = [3,4] # only hi+lo scales
     _set_datasrc(ax, datasrc)
     item = CandlestickItem(ax=ax, datasrc=datasrc, draw_body=draw_body, draw_shadow=draw_shadow, candle_width=candle_width, colorfunc=colorfunc)
-    ax.significant_decimals,ax.significant_eps = datasrc.calc_significant_decimals()
-    item.update_data = partial(_update_data, item)
+    _update_significants(ax, datasrc, force=True)
+    item.update_data = partial(_update_data, None, item)
     ax.addItem(item)
-    # item.setZValue(20)
-    _pre_process_data(item, True)
     return item
 
 
 def volume_ocv(datasrc, candle_width=0.8, ax=None, colorfunc=volume_colorfilter):
     ax = _create_plot(ax=ax, maximize=False)
     datasrc = _create_datasrc(ax, datasrc)
-    if len(datasrc.df.columns) <= 4:
-        datasrc.df.insert(3, '_zero_', [0]*len(datasrc.df)) # base of candles is always zero
-    datasrc.df = datasrc.df.iloc[:,[0,3,4,1,2]] # re-arrange columns for rendering
-    datasrc.scale_cols = [2] # only volume scales Y-zoom
+    _adjust_volume_datasrc(datasrc)
     _set_datasrc(ax, datasrc)
     item = CandlestickItem(ax=ax, datasrc=datasrc, draw_body=True, draw_shadow=False, candle_width=candle_width, colorfunc=colorfunc)
+    _update_significants(ax, datasrc, force=True)
     item.colors['bull_body'] = item.colors['bull_frame']
     if colorfunc == volume_colorfilter: # assume normal volume plot
         item.colors['bull_frame'] = volume_bull_color
         item.colors['bull_body']  = volume_bull_color
         item.colors['bear_frame'] = volume_bear_color
         item.colors['bear_body']  = volume_bear_color
-        ax.vb.y_zoom_offset = 0
+        ax.vb.v_zoom_baseline = 0
     else:
         item.colors['weak_bull_frame'] = brighten(volume_bull_color, 1.2)
         item.colors['weak_bull_body']  = brighten(volume_bull_color, 1.2)
-    item.update_data = partial(_update_data, item)
+    item.update_data = partial(_update_data, _adjust_volume_datasrc, item)
     ax.addItem(item)
     item.setZValue(-1)
-    _pre_process_data(item, True)
     return item
 
 
@@ -910,15 +962,15 @@ def plot(x, y=None, color=None, width=1, ax=None, style=None, legend=None, zooms
     if not zoomscale:
         datasrc.scale_cols = []
     _set_datasrc(ax, datasrc)
-    if legend is not None and ax.legend is None:
-        ax.legend = FinLegendItem(border_color=legend_border_color, fill_color=legend_fill_color, size=None, offset=(3,2))
-        ax.legend.setParentItem(ax.vb)
+    if legend is not None:
+        _create_legend(ax)
+    y = datasrc.y / ax.vb.yscale.scalef
     if style is None or style=='-':
         connect_dots = 'finite' # same as matplotlib; use datasrc.standalone=True if you want to keep separate intervals on a plot
-        item = ax.plot(datasrc.index, datasrc.y, pen=pg.mkPen(used_color, width=width), name=legend, connect=connect_dots)
+        item = ax.plot(datasrc.index, y, pen=pg.mkPen(used_color, width=width), name=legend, connect=connect_dots)
     else:
         symbol = {'v':'t', '^':'t1', '>':'t2', '<':'t3'}.get(style, style) # translate some similar styles
-        ser = datasrc.y.loc[datasrc.y.abs()>=0]
+        ser = y.loc[y.notnull()]
         item = ax.plot(ser.index, ser.values, pen=None, symbol=symbol, symbolPen=None, symbolSize=10, symbolBrush=pg.mkBrush(used_color), name=legend)
         item.scatter._dopaint = item.scatter.paint
         item.scatter.paint = partial(_paint_scatter, item.scatter)
@@ -928,18 +980,13 @@ def plot(x, y=None, color=None, width=1, ax=None, style=None, legend=None, zooms
     item.opts['handed_color'] = color
     item.ax = ax
     item.datasrc = datasrc
-    # check if no epsilon set yet
-    if 0.99 < ax.significant_eps/significant_eps < 1.01:
-        try:
-            ax.significant_decimals,ax.significant_eps = datasrc.calc_significant_decimals()
-        except:
-            pass # probably full av NaNs
-    item.update_data = partial(_update_data, item)
-    _pre_process_data(item, zoomscale)
+    _update_significants(ax, datasrc, force=False)
+    item.update_data = partial(_update_data, None, item)
     if ax.legend is not None:
         for _,label in ax.legend.items:
-            label.setAttr('justify', 'left')
-            label.setText(label.text, color=legend_text_color)
+            if label.text == legend:
+                label.setAttr('justify', 'left')
+                label.setText(label.text, color=legend_text_color)
     return item
 
 
@@ -950,9 +997,28 @@ def labels(x, y=None, labels=None, color=None, ax=None, anchor=(0.5,1)):
     datasrc.scale_cols = [] # don't use this for scaling
     _set_datasrc(ax, datasrc)
     item = ScatterLabelItem(ax=ax, datasrc=datasrc, color=color, anchor=anchor)
-    item.update_data = partial(_update_data, item)
+    _update_significants(ax, datasrc, force=False)
+    item.update_data = partial(_update_data, None, item)
     ax.addItem(item)
-    _pre_process_data(item, False)
+    if ax.vb.v_zoom_scale > 0.9: # adjust to make hi/lo text fit
+        ax.vb.v_zoom_scale = 0.9
+    return item
+
+
+def add_legend(text, ax=None):
+    ax = _create_plot(ax=ax, maximize=False)
+    _create_legend(ax)
+    row = ax.legend.layout.rowCount()
+    label = pg.LabelItem(text, color=legend_text_color, justify='left')
+    ax.legend.layout.addItem(label, row, 0, 1, 2)
+    return label
+
+
+def fill_between(plot0, plot1, color=None):
+    used_color = brighten(_get_color(plot0.ax, None, color), 1.3)
+    item = pg.FillBetweenItem(plot0, plot1, brush=pg.mkBrush(used_color))
+    item.ax = plot0.ax
+    plot0.ax.addItem(item)
     return item
 
 
@@ -973,7 +1039,7 @@ def set_y_range(ymin, ymax, ax=None):
 def set_yscale(yscale='linear', ax=None):
     ax = _create_plot(ax=ax, maximize=False)
     ax.setLogMode(y=(yscale=='log'))
-    ax.vb.yscale = yscale
+    ax.vb.yscale = YScale(yscale, ax.vb.yscale.scalef)
 
 
 def add_band(y0, y1, color=band_color, ax=None):
@@ -1023,15 +1089,18 @@ def remove_text(text):
     text.ax.removeItem(text)
 
 
-def set_time_inspector(inspector, ax=None):
-    '''Callback when clicked like so: inspector().'''
+def set_time_inspector(inspector, ax=None, when='click'):
+    '''Callback when clicked like so: inspector(x, y).'''
     ax = ax if ax else last_ax
     win = ax.vb.win
-    win.proxy_click = pg.SignalProxy(win.scene().sigMouseClicked, slot=partial(_time_clicked, ax, inspector))
+    if when == 'hover':
+        win.proxy_hover = pg.SignalProxy(win.scene().sigMouseMoved, rateLimit=15, slot=partial(_inspect_pos, ax, inspector))
+    else:
+        win.proxy_click = pg.SignalProxy(win.scene().sigMouseClicked, slot=partial(_inspect_clicked, ax, inspector))
 
 
 def add_crosshair_info(infofunc, ax=None):
-    '''Callback when crosshair updated like so: info(x,y,xtext,ytext); the info()
+    '''Callback when crosshair updated like so: info(ax,x,y,xtext,ytext); the info()
        callback must return two values: xtext and ytext.'''
     ax = _create_plot(ax=ax, maximize=False)
     ax.crosshair.infos.append(infofunc)
@@ -1055,9 +1124,17 @@ def autoviewrestore(enable=True):
 
 
 def show():
-    if viewrestore:
-        for win in windows:
-            _loadwindata(win)
+    for win in windows:
+        vbs = set(ax.vb for ax in win.ci.items)
+        for vb in vbs:
+            _pre_process_data(vb)
+        if viewrestore:
+            if _loadwindata(win):
+                continue
+        for vb in vbs:
+            if vb.datasrc:
+                vb.update_y_zoom(vb.datasrc.init_x0, vb.datasrc.init_x1)
+                break
     _repaint_candles()
     if windows:
         QtGui.QApplication.instance().exec_()
@@ -1084,13 +1161,17 @@ def _loadwindata(win):
     except:
         return
     kvs = {k:v for k,v in settings}
-    for ax in win.ci.items:
-        ds = ax.vb.datasrc
+    vbs = set(ax.vb for ax in win.ci.items)
+    for vb in vbs:
+        ds = vb.datasrc
         if ds:
             period = ds.period
             if kvs['min_x'] >= ds.x.iloc[0]-period and kvs['max_x'] <= ds.x.iloc[-1]+period:
                 x0,x1 = ds.x.loc[ds.x>=kvs['min_x']].index[0], ds.x.loc[ds.x<=kvs['max_x']].index[-1]
-                ax.vb.update_y_zoom(x0, x1)
+                if x1 == len(ds.x)-1:
+                    x1 += right_margin_candles
+                vb.update_y_zoom(x0, x1)
+    return True
 
 
 def _savewindata(win):
@@ -1129,15 +1210,19 @@ def _add_timestamp_plot(win, prev_ax, viewbox, index, yscale):
     if prev_ax is not None:
         prev_ax.hideAxis('bottom') # hide the whole previous axis
         win.nextRow()
-    ax = pg.PlotItem(viewBox=viewbox, axisItems={'bottom': EpochAxisItem(vb=viewbox, orientation='bottom')}, name='plot-%i'%index)
+    axes = {'bottom': EpochAxisItem(vb=viewbox, orientation='bottom'),
+            'left':   YAxisItem(vb=viewbox, orientation='left')}
+    ax = pg.PlotItem(viewBox=viewbox, axisItems=axes, name='plot-%i'%index)
     ax.axes['left']['item'].textWidth = y_label_width # this is to put all graphs on equal footing when texts vary from 0.4 to 2000000
     ax.axes['left']['item'].setStyle(tickLength=-5) # some bug, totally unexplicable (why setting the default value again would fix repaint width as axis scale down)
     ax.axes['left']['item'].setZValue(30) # put axis in front instead of behind data
     ax.axes['bottom']['item'].setZValue(30)
-    ax.setLogMode(y=(yscale=='log'))
+    ax.setLogMode(y=(yscale.scaletype=='log'))
     ax.significant_decimals = significant_decimals
     ax.significant_eps = significant_eps
     ax.crosshair = FinCrossHair(ax, color=cross_hair_color)
+    ax.hideButtons()
+    ax.overlay = partial(_overlay, ax)
     if index%2:
         viewbox.setBackgroundColor(odd_plot_background)
     viewbox.setParent(ax)
@@ -1145,8 +1230,50 @@ def _add_timestamp_plot(win, prev_ax, viewbox, index, yscale):
     return ax
 
 
+def _overlay(ax, scale=0.25):
+    global overlay_axs
+    viewbox = FinViewBox(ax.vb.win, init_steps=ax.vb.init_steps, yscale=YScale('linear', 1))
+    viewbox.v_zoom_scale = scale
+    ax.vb.win.centralWidget.scene().addItem(viewbox)
+    viewbox.setXLink(ax.vb)
+    def updateView():
+        viewbox.setGeometry(ax.vb.sceneBoundingRect())
+    axo = pg.PlotItem()
+    axo.significant_decimals = significant_decimals
+    axo.significant_eps = significant_eps
+    axo.vb = viewbox
+    axo.hideAxis('left')
+    axo.hideAxis('bottom')
+    axo.hideButtons()
+    viewbox.addItem(axo)
+    ax.vb.sigResized.connect(updateView)
+    overlay_axs.append(axo)
+    return axo
+
+
+def _create_legend(ax):
+    if ax.legend is None:
+        ax.legend = FinLegendItem(border_color=legend_border_color, fill_color=legend_fill_color, size=None, offset=(3,2))
+        ax.legend.setParentItem(ax.vb)
+
+
+def _update_significants(ax, datasrc, force):
+    # check if no epsilon set yet
+    default_dec = 0.99 < ax.significant_decimals/significant_decimals < 1.01
+    default_eps = 0.99 < ax.significant_eps/significant_eps < 1.01
+    if force or (default_dec and default_eps):
+        try:
+            sd,se = datasrc.calc_significant_decimals()
+            if default_dec or sd > ax.significant_decimals:
+                ax.significant_decimals = sd
+            if default_eps or se < ax.significant_eps:
+                ax.significant_eps = se
+        except:
+            pass # datasrc probably full av NaNs
+
+
 def _is_standalone(timeser):
-    # more than N percent gaps or reversals probably means this is a standalone plot
+    # more than N percent gaps or time reversals probably means this is a standalone plot
     return timeser.isnull().sum() + (timeser.diff()<=0).sum() > len(timeser)*0.1
 
 
@@ -1154,7 +1281,7 @@ def _create_series(a):
     return a if isinstance(a, pd.Series) else pd.Series(a)
 
 
-def _create_datasrc(ax, *args):
+def _create_datasrc(ax, *args, datacols=1):
     def do_create(*args):
         args = [a for a in args if a is not None]
         if len(args) == 1 and type(args[0]) == PandasDataSource:
@@ -1169,14 +1296,18 @@ def _create_datasrc(ax, *args):
         return PandasDataSource(pd.concat(args, axis=1))
     datasrc = do_create(*args)
     # check if time column missing
-    if len(datasrc.df.columns) == 1:
+    if len(datasrc.df.columns) == datacols:
         # assume time data has already been added before
         for a in ax.vb.win.ci.items:
             if a.vb.datasrc and len(a.vb.datasrc.df.columns) >= 2:
+                datasrc.df.columns = a.vb.datasrc.df.columns[1:len(datasrc.df.columns)+1]
                 col = a.vb.datasrc.df.columns[0]
                 datasrc.df.insert(0, col, a.vb.datasrc.df[col])
                 datasrc = PandasDataSource(datasrc.df)
                 break
+    # FIX: stupid QT bug causes rectangles larger than 2G to flicker, so scale rendering down some
+    if datasrc.df.iloc[:, 1:].max(numeric_only=True).max() > 1e8: # too close to 2G for comfort
+        ax.vb.yscale.set_scale(int(1e8))
     return datasrc
 
 
@@ -1203,8 +1334,17 @@ def _has_timecol(df):
     return len(df.columns) >= 2
 
 
-def _update_data(item, ds):
+def _adjust_volume_datasrc(datasrc):
+    if len(datasrc.df.columns) <= 4:
+        datasrc.df.insert(3, '_zero_', [0]*len(datasrc.df)) # base of candles is always zero
+    datasrc.df = datasrc.df.iloc[:,[0,3,4,1,2]] # re-arrange columns for rendering
+    datasrc.scale_cols = [1, 2] # scale by both baseline and volume
+
+
+def _update_data(adjustfunc, item, ds):
     ds = _create_datasrc(item.ax, ds)
+    if adjustfunc:
+        adjustfunc(ds)
     item.datasrc.update(ds)
     if isinstance(item, FinPlotItem):
         item.dirty = True
@@ -1216,19 +1356,20 @@ def _update_data(item, ds):
     x0 = x1 - tr.width()
     for ax in item.ax.vb.win.ci.items:
         ax.setLimits(xMin=x_min, xMax=x1)
-    if tr.right() >= x1-5:
+    if tr.right() >= x1 - 5 - 2*right_margin_candles:
         for ax in item.ax.vb.win.ci.items:
             ax.vb.update_y_zoom(x0, x1)
     for ax in item.ax.vb.win.ci.items:
         ax.vb.update()
 
 
-def _pre_process_data(item, zoomscale):
-    if zoomscale:
-        item.ax.vb.y_max = np.nanmax(item.datasrc.y)
-        item.ax.vb.y_min = np.nanmin(item.datasrc.y)
-        if item.ax.vb.y_min <= 0:
-            item.ax.vb.y_positive = False
+def _pre_process_data(vb):
+    if vb.datasrc and vb.datasrc.scale_cols:
+        df = vb.datasrc.df.iloc[:, vb.datasrc.scale_cols]
+        vb.y_max = df.max().max()
+        vb.y_min = df.min().min()
+        if vb.y_min <= 0:
+            vb.y_positive = False
 
 
 def _set_plot_x_axis_leader(ax):
@@ -1243,17 +1384,18 @@ def _set_plot_x_axis_leader(ax):
 
 def _set_x_limits(ax, datasrc):
     x0 = -0.5
-    ax.setLimits(xMin=x0, xMax=datasrc.init_x1)
-    return x0, datasrc.init_x1
+    x1 = datasrc.init_x1 + right_margin_candles # add another margin to get the "snap back" sensation
+    ax.setLimits(xMin=x0, xMax=x1)
+    return x0, x1
 
 
 def _repaint_candles():
     '''Candles are only partially drawn, and therefore needs manual dirty reminder whenever it goes off-screen.'''
-    for win in windows:
-        for ax in win.ci.items:
-            for item in ax.items:
-                if isinstance(item, FinPlotItem):
-                    item.repaint()
+    axs = [ax for win in windows for ax in win.ci.items] + overlay_axs
+    for ax in axs:
+        for item in ax.items:
+            if isinstance(item, FinPlotItem):
+                item.repaint()
 
 
 def _paint_scatter(item, p, *args):
@@ -1291,6 +1433,16 @@ def _key_pressed(vb, ev):
     return True
 
 
+def _mouse_clicked(vb, ev):
+    if ev.button() == 8: # back
+        vb.pan_x(percent=-30)
+    elif ev.button() == 16: # fwd
+        vb.pan_x(percent=+30)
+    else:
+        return False
+    return True
+
+
 def _mouse_moved(win, evs):
     if not evs:
         evs = win._last_mouse_evs
@@ -1318,14 +1470,24 @@ def _wheel_event_wrapper(self, orig_func, ev):
     orig_func(self, ev)
 
 
-def _time_clicked(ax, inspector, evs):
+def _inspect_clicked(ax, inspector, evs):
     if evs[-1].accepted:
         return
     pos = evs[-1].scenePos()
-    point = ax.vb.mapSceneToView(pos)
-    t = point.x() - 0.5
-    t = ax.vb.datasrc.closest_time(t)
-    inspector(t, point.y())
+    return _inspect_pos(ax, inspector, (pos,))
+
+
+def _inspect_pos(ax, inspector, poss):
+    point = ax.vb.mapSceneToView(poss[-1])
+    t = point.x() + 0.5
+    try:
+        t = ax.vb.datasrc.closest_time(t)
+    except KeyError: # when clicking beyond right_margin_candles
+        return
+    try:
+        inspector(t, point.y())
+    except Exception as e:
+        print(type(e), e)
 
 
 def brighten(color, f):
@@ -1350,21 +1512,33 @@ def _get_color(ax, style, wanted_color):
 def _pdtime2epoch(t):
     if isinstance(t, pd.Series):
         if isinstance(t.iloc[0], pd.Timestamp):
-            return t.astype('int64') // int(1e9)
-        if np.nanmax(t.values) > 1e10: # handle ms epochs
+            return t.astype('int64') // int(1e6)
+        if np.nanmax(t.values) > 1e13: # handle ns epochs
             return t.astype('float64') / 1e3
+        if np.nanmax(t.values) < 1e10: # handle s epochs
+            return t.astype('float64') * 1e3
     return t
 
 
 def _pdtime2index(ax, ts):
     if isinstance(ts.iloc[0], pd.Timestamp):
-        ts = ts.astype('int64') // int(1e9)
-    elif np.nanmax(ts.values) > 1e10: # handle ms epochs
+        ts = ts.astype('int64') // int(1e6)
+    elif np.nanmax(ts.values) > 1e13: # handle ns epochs
         ts = ts.astype('float64') / 1e3
+    elif np.nanmax(ts.values) < 1e10: # handle s epochs
+        ts = ts.astype('float64') * 1e3
     r = []
+    datasrc = _get_datasrc(ax)
     for t in ts:
-        xs = ax.vb.datasrc.x
-        i1 = xs.loc[xs>t].index[0]
+        xs = datasrc.x
+        xss = xs.loc[xs>t]
+        if len(xss) == 0:
+            t0 = xs.iloc[-1]
+            if t0 == t:
+                r.append(len(xs)-1)
+                continue
+            assert t <= t0, 'must plot this primitive in prior time-range'
+        i1 = xss.index[0]
         i0 = i1-1
         t0,t1 = xs.loc[i0], xs.loc[i1]
         dt = (t-t0) / (t1-t0)
@@ -1372,17 +1546,40 @@ def _pdtime2index(ax, ts):
     return r
 
 
-def _epoch2local(datasrc, x, strip_seconds=True):
+def _get_datasrc(ax):
+    if ax.vb.datasrc is not None:
+        return ax.vb.datasrc
+    vbs = set(ax.vb for win in windows for ax in win.ci.items)
+    for vb in vbs:
+        if vb.datasrc:
+            return vb.datasrc
+    assert ax.vb.datasrc, 'not possible to plot this primitive without a prior time-range to compare to'
+
+
+def _epoch2local(datasrc, x):
+    if not datasrc:
+        return ''
     try:
         x += 0.5
         t,_,_,_,cnt = datasrc.hilo(x, x)
         if cnt:
-            s = datetime.fromtimestamp(t).isoformat().replace('T',' ')
-            if epoch_period >= 24*60*60 and strip_seconds:
-                return s.partition(' ')[0] # might as well strip time as well
-            return s.rpartition(':' if strip_seconds or '.' not in s else '.')[0]
-    except:
-        pass
+            s = datetime.fromtimestamp(t/1000).isoformat().replace('T',' ')
+            if epoch_period >= 24*60*60:
+                i = s.index(' ')
+            elif epoch_period >= 60*60:
+                i = s.index(':')
+            elif epoch_period >= 60:
+                i = s.rindex(':')
+            elif epoch_period >= 1:
+                i = s.index('.') if '.' in s else len(s)
+            elif epoch_period >= 0.001:
+                i = -3
+            else:
+                i = len(s)
+            return s[:i]
+    except Exception as e:
+        import traceback
+        traceback.print_exc()
     return ''
 
 
@@ -1390,9 +1587,9 @@ def _round_to_significant(rng, rngmax, x, significant_decimals, significant_eps)
     is_highres = rng/significant_eps > 1e2 and (rngmax>1e7 or rngmax<1e-2)
     sd = significant_decimals
     if is_highres and abs(x)>0:
-        exp10 = floor(log10(abs(x)))
+        exp10 = floor(np.log10(abs(x)))
         x = x / (10**exp10)
-        sd = min(5, sd+int(log10(rngmax)))
+        sd = min(5, sd+int(np.log10(rngmax)))
         fmt = '%%%i.%ife%%i' % (sd, sd)
         r = fmt % (x, exp10)
     else:
@@ -1414,14 +1611,13 @@ def _roihandle_move_snap(vb, orig_func, pos, modifiers=QtCore.Qt.KeyboardModifie
 
 
 def _clamp_xy(ax, x, y):
-    if ax.vb.yscale == 'log':
-        y = 10**y
+    y = ax.vb.yscale.xform(y)
     if clamp_grid:
         x = round(x)
         eps = ax.significant_eps
-        y -= fmod(y+eps*0.5, eps) - eps*0.5
-    if ax.vb.yscale == 'log':
-        y = np.log10(y)
+        eps2 = np.sign(y) * 0.5 * eps
+        y -= fmod(y+eps2, eps) - eps2
+    y = ax.vb.yscale.invxform(y, verify=True)
     return x, y
 
 
@@ -1434,10 +1630,18 @@ def _clamp_point(ax, p):
 
 def _draw_line_segment_text(polyline, segment, pos0, pos1):
     diff = pos1 - pos0
-    mins = int(abs(diff.x()*epoch_period) // 60)
+    fsecs = abs(diff.x()*epoch_period)
+    secs = int(fsecs)
+    mins = secs//60
     hours = mins//60
     mins = mins%60
-    if hours < 24:
+    secs = secs%60
+    if hours==0 and mins==0 and secs < 60 and epoch_period < 1:
+        msecs = int((fsecs-int(fsecs))*1000)
+        ts = '%0.2i:%0.2i.%0.3i' % (mins, secs, msecs)
+    elif hours==0 and mins < 60 and epoch_period < 60:
+        ts = '%0.2i:%0.2i:%0.2i' % (hours, mins, secs)
+    elif hours < 24:
         ts = '%0.2i:%0.2i' % (hours, mins)
     else:
         days = hours // 24
@@ -1445,11 +1649,12 @@ def _draw_line_segment_text(polyline, segment, pos0, pos1):
         ts = '%id %0.2i:%0.2i' % (days, hours, mins)
         if ts.endswith(' 00:00'):
             ts = ts.partition(' ')[0]
+    ysc = polyline.vb.yscale
     if polyline.vb.y_positive:
-        y0,y1 = (10**pos0.y(),10**pos1.y()) if polyline.vb.yscale == 'log' else (pos0.y(),pos1.y())
+        y0,y1 = ysc.xform(pos0.y()), ysc.xform(pos1.y())
         value = '%+.2f %%' % (100 * y1 / y0 - 100)
     else:
-        dy = diff.y()
+        dy = ysc.xform(diff.y())
         if dy and (abs(dy) >= 1e4 or abs(dy) <= 1e-2):
             value = '%+3.3g' % dy
         else:
@@ -1467,7 +1672,7 @@ def _draw_line_extra_text(polyline, segment, pos0, pos1):
             h1 = prev_text.segment.handles[1]['item']
             prev_change = h1.pos().y() - h0.pos().y()
             this_change = pos1.y() - pos0.y()
-            if not abs(prev_change) > 1e-8:
+            if not abs(prev_change) > 1e-14:
                 break
             change_part = abs(this_change / prev_change)
             return ' = 1:%.2f ' % change_part
@@ -1475,6 +1680,15 @@ def _draw_line_extra_text(polyline, segment, pos0, pos1):
     return ''
 
 
+try:
+    qtver = '%d.%d' % (QtCore.QT_VERSION//256//256, QtCore.QT_VERSION//256%256)
+    if qtver not in ('5.9', '5.13'):
+        print('WARNING: your version of Qt may not plot curves containing NaNs and is not recommended.')
+        print('See https://github.com/pyqtgraph/pyqtgraph/issues/1057')
+except:
+    pass
+
+
 # default to black-on-white
 pg.widgets.GraphicsView.GraphicsView.wheelEvent = partialmethod(_wheel_event_wrapper, pg.widgets.GraphicsView.GraphicsView.wheelEvent)
 # pick up win resolution
